# Copyright (C) 2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

# * The software is provided "as is" without any warranty, neither express
#   nor implied.
# * In no event will the author be held liable for any damage(s) arising
#   from the use of the software.
# * Redistribution of the software is allowed only in unmodified form.
# * Permission is granted to use the software solely for personal private
#   and non-commercial purposes.
# * An individuals use of the software in his or her capacity or function
#   as an agent, (independent) contractor, employee, member or officer of
#   a business, corporation or organization (commercial or non-commercial)
#   does not qualify as personal private and non-commercial purpose.
# * Without written approval from the author the software must not be used
#   for a business, for commercial, corporate, governmental, military or
#   organizational purposes of any kind, or in a commercial, corporate,
#   governmental, military or organizational environment of any kind.

.SUFFIXES:

all: TLS_DEMO.DOS TLS_DEMO.COM TLS_DEMO.DLL TLS_DEMO.EXE
	"$(MAKEDIR)\TLS_DEMO.COM"
	"$(MAKEDIR)\TLS_DEMO.EXE"
	CONTROL.EXE "$(MAKEDIR)\TLS_DEMO.DLL"
	REGSVR32.EXE /I:"$**" "$(MAKEDIR)\TLS_DEMO.DLL"
	REGSVR32.EXE "$(MAKEDIR)\TLS_DEMO.DLL"
	REGSVR32.EXE /U "$(MAKEDIR)\TLS_DEMO.DLL"
	RUNDLL32.EXE "$(MAKEDIR)\TLS_DEMO.DLL",#1 $?

clean:
	ERASE TLS_DEMO.DOS TLS_DEMO.EXP TLS_DEMO.LIB TLS_DEMO.OBJ TLS_DEMO.TMP

clobber: clean
	ERASE TLS_DEMO.COM TLS_DEMO.DLL TLS_DEMO.EXE

TLS_DEMO.DOS: TLS_DEMO.MAK
	CERTUTIL.EXE /DecodeHex /F /V << $@
4D 5A 90 00 01 00 00 00 04 00 00 00 FF FF 00 00   MZ..............
E0 00 00 00 43 00 00 00 40 00 00 00 00 00 00 00   ....C...@.......
00 00 00 00 19 57 04 27 00 00 00 00 00 00 00 00   .....W.'........
00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00   ................
28 43 29 6F 70 79 72 69 67 68 74 20 32 30 30 34   (C)opyright 2004
2D 32 30 32 34 2C 20 53 74 65 66 61 6E 20 4B 61   -2024, Stefan Ka
6E 74 68 61 6B 20 3C 73 74 65 66 61 6E 2E 6B 61   nthak <stefan.ka
6E 74 68 61 6B 40 6E 65 78 67 6F 2E 64 65 3E 0D   nthak@nexgo.de>.
0A 07 24 0E 1F 33 D2 B4 09 CD 21 B8 01 4C CD 21   .....3....!..L.!
<<NOKEEP

TLS_DEMO.OMF: TLS_DEMO.MAK
	ML.EXE /Brepro /Fo$@ /nologo /omf /Ta<< /W3 /X
; Copyright (C) 2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

	.model	tiny

e_header struct				; DOS .EXE header
e_magic	word	5A4Dh			; magic number "MZ"
e_cblp	word	stop - header		; bytes on last page of file
e_cp	word	1			; pages in file
e_crlc	word	0			; number of relocation table entries
e_cparhdr word	sizeof header / 16	; size of header in paragraphs
e_minalloc word	0			; minimum extra paragraphs needed
e_maxalloc word	65535			; maximum extra paragraphs needed
e_ss	word	0			; initial (PSP relative) SS value
e_sp	word	stack - header		; initial (SS relative) SP value
e_csum	word	0			; checksum
e_ip	word	sizeof message		; initial (CS relative) IP value
e_cs	word	0			; initial (PSP relative) CS value
e_lfarlc word	sizeof header		; file address of relocation table
e_ovno	word	0			; overlay number
e_res	word	4 dup (0)		; reserved words
e_oemid	word	5719h			; OEM identifier
e_oeminfo word	2704h			; OEM information
e_res2	word	10 dup (0)		; reserved words
e_lfanew dword	0			; file address of new exe header
e_header ends

	.code

header	e_header <>

message	byte	"(C)opyright 2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>", 0Dh, 0Ah, 07h, 24h

start:
	push	cs
	pop	ds
	xor	dx, dx
	mov	ah, 09h
	int	21h

	mov	ax, 4C01h
	int	21h
stop:
	dword	20 dup (?)
stack:
	end	start
<<NOKEEP

TLS_DEMO.COM: TLS_DEMO.MAK
	CL.EXE /c /Fo$*.TMP /GAF /Gwy /Oxy /Tc<< /W4 /we4013 /Zl
// Copyright (C) 2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

// https://msdn.microsoft.com/en-us/library/7f0aews7.aspx

#pragma comment(compiler)

#define STRICT
#define UNICODE
#define WIN32_LEAN_AND_MEAN

#include <windows.h>

extern	const	IMAGE_DOS_HEADER	__ImageBase;

// https://msdn.microsoft.com/en-us/library/gg463119.aspx
// https://msdn.microsoft.com/en-us/library/ms680547.aspx

const	LPCWSTR	szReason[4] = {L"process detach",
		               L"process attach",
		               L"thread attach",
		               L"thread detach"};

__declspec(safebuffers)
VOID	WINAPI	TLSConsole(HMODULE hModule, DWORD dwReason, LPVOID lpUnused)
{
	WCHAR	szOutput[1024];

	// https://msdn.microsoft.com/en-us/library/ms647550.aspx

	INT	niOutput = wsprintf(szOutput, L"\n"
		                              __LPREFIX(__FUNCTION__) L"() Function\n"
		                              L"\n"
		                              L"Return address @ 0x%p = 0x%p\n"
		                              L"\n"
		                              L"Arguments:\n"
		                              L"\tModule = 0x%p\n"
		                              L"\tReason = %u (%s)\n"
		                              L"\tUnused = 0x%p\n",
		                    _AddressOfReturnAddress(), _ReturnAddress(),
		                    hModule, dwReason, szReason[dwReason], lpUnused);

	if (niOutput > 0)

		// https://msdn.microsoft.com/en-us/library/ms687401.aspx
		// https://msdn.microsoft.com/en-us/library/ms683231.aspx

		WriteConsole(GetStdHandle(STD_ERROR_HANDLE), szOutput, niOutput, &dwReason, NULL);
}

DWORD	_tls_index = ~0UL;	// NOTE: updated by the module loader!

const	PIMAGE_TLS_CALLBACK	_tls_callbacks[] = {TLSConsole, NULL};

// BUG: the module loader does NOT support the 'SizeOfZeroFill' member!

const	IMAGE_TLS_DIRECTORY	_tls_used = {&__ImageBase,
				             &__ImageBase + 5 * 4096 / sizeof(IMAGE_DOS_HEADER),
				             &_tls_index,
				             _tls_callbacks,
				             ~0UL,
				             0UL};

// https://msdn.microsoft.com/en-us/library/f9t8842e.aspx

__declspec(noreturn)
VOID	WINAPI	wmainCRTStartup(VOID)
{
	WCHAR	szOutput[1024];
	INT	niOutput;

	// https://msdn.microsoft.com/en-us/library/ms647550.aspx

	niOutput = wsprintf(szOutput, L"\n"
	                              __LPREFIX(__FUNCTION__) L"() Function\n"
	                              L"\n"
	                              L"Return address @ 0x%p = 0x%p\n"
	                              L"\n"
	                              L"TLS index = %ld\n"
	                              L"TLS value = 0x%p\n"
	                              L"TLS array @ 0x%p\n"
	                              L"TLS block @ 0x%p\n",
	                    _AddressOfReturnAddress(), _ReturnAddress(),
	                    _tls_index,
	                    TlsGetValue(_tls_index),
#ifdef _M_IX86
	                    __readfsdword(44),
	                    ((LPVOID *) __readfsdword(44))[_tls_index]);
#elif _M_AMD64
	                    __readgsqword(88),
	                    ((LPVOID *) __readgsqword(88))[_tls_index]);
#else
#error
#endif
	if (niOutput > 0)

		// https://msdn.microsoft.com/en-us/library/ms687401.aspx
		// https://msdn.microsoft.com/en-us/library/ms683231.aspx

		WriteConsole(GetStdHandle(STD_ERROR_HANDLE), szOutput, niOutput, &niOutput, NULL);

	// https://msdn.microsoft.com/en-us/library/ms682658.aspx

	ExitProcess(_tls_index);
}
<<NOKEEP
	LINK.EXE /LINK /BREPRO /DEFAULTLIB:KERNEL32.LIB /DEFAULTLIB:USER32.LIB /DYNAMICBASE /ENTRY:wmainCRTStartup /LARGEADDRESSAWARE /LAST:.data /MERGE:.rdata=.const /MERGE:.text=.code /NOCOFFGRPINFO /NOLOGO /NXCOMPAT /OPT:REF /OSVERSION:5.0 /OUT:$@ /RELEASE /STUB:TLS_DEMO.DOS /SUBSYSTEM:CONSOLE /SWAPRUN:CD,NET /TEST /VERSION:0.815 $*.TMP

TLS_DEMO.DLL: TLS_DEMO.OBJ
	LINK.EXE /LINK /BREPRO /DEFAULTLIB:KERNEL32.LIB /DEFAULTLIB:USER32.LIB /DLL /DYNAMICBASE /ENTRY:_DllMainCRTStartup /EXPORT:DllInstall,PRIVATE /EXPORT:DllRegisterServer,PRIVATE /EXPORT:DllUnregisterServer,PRIVATE /EXPORT:RunDll,@1,NONAME,PRIVATE /LARGEADDRESSAWARE /LAST:.data /MERGE:.rdata=.const /MERGE:.text=.code /NOCOFFGRPINFO /NOLOGO /NXCOMPAT /OPT:REF /OSVERSION:5.0 /OUT:$@ /RELEASE /STUB:TLS_DEMO.DOS /SUBSYSTEM:WINDOWS /SWAPRUN:CD,NET /TEST /VERSION:0.815 $*

TLS_DEMO.EXE: TLS_DEMO.OBJ
	LINK.EXE /LINK /BREPRO /DEFAULTLIB:KERNEL32.LIB /DEFAULTLIB:USER32.LIB /DYNAMICBASE /ENTRY:wWinMainCRTStartup /LARGEADDRESSAWARE /LAST:.data /MERGE:.rdata=.const /MERGE:.text=.code /NOCOFFGRPINFO /NOLOGO /NXCOMPAT /OPT:REF /OSVERSION:5.0 /OUT:$@ /RELEASE /STUB:TLS_DEMO.DOS /SUBSYSTEM:WINDOWS /SWAPRUN:CD,NET /TEST /VERSION:0.815 $*

TLS_DEMO.OBJ: TLS_DEMO.MAK
	CL.EXE /c /Fo$@ /GAF /Gwy /Oxy /Tc<< /W4 /we4013 /Zl
// Copyright (C) 2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

// https://msdn.microsoft.com/en-us/library/7f0aews7.aspx

#pragma comment(compiler)

#define STRICT
#define UNICODE
#define WIN32_LEAN_AND_MEAN

#include <windows.h>

#pragma intrinsic(_AddressOfReturnAddress, _ReturnAddress)

const	LPCWSTR	szReason[4] = {L"process detach",
		               L"process attach",
		               L"thread attach",
		               L"thread detach"};

__declspec(safebuffers)
VOID	WINAPI	TLSWindows(HMODULE hModule, DWORD dwReason, LPVOID lpUnused)
{
	WCHAR	szOutput[1024];

	// https://msdn.microsoft.com/en-us/library/ms647550.aspx

	INT	niOutput = wsprintf(szOutput, L"Return address @ 0x%p = 0x%p\n"
		                              L"\n"
		                              L"Arguments:\n"
		                              L"\tModule = 0x%p\n"
		                              L"\tReason = %u (%s)\n"
		                              L"\tUnused = 0x%p\n",
		                    _AddressOfReturnAddress(), _ReturnAddress(),
		                    hModule, dwReason, szReason[dwReason], lpUnused);

	if (niOutput > 0)
	{
	//	szOutput[niOutput] = L'\0';

		// https://msdn.microsoft.com/en-us/library/aa363362.aspx

		OutputDebugStringA(__FUNCTION__ "() Function");

		// https://msdn.microsoft.com/en-us/library/ms645507.aspx

		MessageBoxEx(HWND_DESKTOP,
		             szOutput,
		             __LPREFIX(__FUNCTION__) L"() Function",
		             MB_OK,
		             MAKELANGID(LANG_ENGLISH, SUBLANG_NEUTRAL));
	}
}

#if 0
// https://msdn.microsoft.com/en-us/library/6yh4a9k1.aspx

// Offset of 'ThreadLocalStoragePointer' member in TEB (Thread Environment Block):
//  44 for I386, referenced in compiled code, 88 for AMD64, not referenced!

extern	OFFSET	_tls_array;

// https://msdn.microsoft.com/en-us/library/thfhx4st.aspx

#pragma data_seg(".tls")

BYTE	*_tls_start = NULL;

#pragma data_seg(".tls$$~~~")

BYTE	*_tls_end = NULL;

#pragma data_seg()

// https://msdn.microsoft.com/en-us/library/7f0aews7.aspx
// https://msdn.microsoft.com/en-us/library/sf9b18xk.aspx

#pragma comment(linker, "/SECTION:.tls,r")
#endif

DWORD	_tls_index = ~0UL;	// NOTE: assigned by the module loader!

// https://msdn.microsoft.com/en-us/library/1dc22465.aspx

const	PIMAGE_TLS_CALLBACK	_tls_callbacks[] = {TLSWindows, NULL};

// BUG: the module loader does NOT support the 'SizeOfZeroFill' member!

#pragma const_seg(".rdata$$T")

const	IMAGE_TLS_DIRECTORY	_tls_used = {0UL,
				             0UL,
				             &_tls_index,
				             _tls_callbacks,
				             'VOID',
				             0UL};
#pragma const_seg()

// https://support.microsoft.com/en-us/kb/249873
//
// https://msdn.microsoft.com/en-us/library/bb759846.aspx
//
// REGSVR32.EXE /I[:[<command line>]] [/N] [/S] [/U] TLS_DEMO.DLL

__declspec(safebuffers)
HRESULT	WINAPI	DllInstall(BOOL bInstall, LPCWSTR lpszCmdLine)
{
	WCHAR	szOutput[1024];

	// https://msdn.microsoft.com/en-us/library/ms647550.aspx

	INT	niOutput = wsprintf(szOutput, L"Return address @ 0x%p = 0x%p\n"
		                              L"\n"
		                              L"Arguments:\n"
		                              L"\tInstall = %d\n"
		                              L"\tCmdLine = %s\n",
		                    _AddressOfReturnAddress(), _ReturnAddress(),
		                    bInstall, lpszCmdLine);
	INT	niButton = 0;

	if (niOutput <= 0)
		return S_FALSE;

//	szOutput[niOutput] = L'\0';

	// https://msdn.microsoft.com/en-us/library/ms645507.aspx

	niButton = MessageBoxEx(HWND_DESKTOP,
	                        szOutput,
	                        __LPREFIX(__FUNCTION__) L"() Function",
	                        MB_OK,
	                        MAKELANGID(LANG_ENGLISH, SUBLANG_NEUTRAL));

	// https://msdn.microsoft.com/en-us/library/ms688560.aspx

	return niButton == IDOK ? S_OK : S_FALSE;
}

// https://msdn.microsoft.com/en-us/library/ms682162.aspx
//
// REGSVR32.EXE [/S] TLS_DEMO.DLL

__declspec(safebuffers)
HRESULT	WINAPI	DllRegisterServer(VOID)
{
	WCHAR	szOutput[1024];

	// https://msdn.microsoft.com/en-us/library/ms647550.aspx

	INT	niOutput = wsprintf(szOutput, L"Return address @ 0x%p = 0x%p\n",
		                    _AddressOfReturnAddress(), _ReturnAddress());
	INT	niButton = 0;

	if (niOutput <= 0)
		return S_FALSE;

//	szOutput[niOutput] = L'\0';

	// https://msdn.microsoft.com/en-us/library/ms645507.aspx

	niButton = MessageBoxEx(HWND_DESKTOP,
	                        szOutput,
	                        __LPREFIX(__FUNCTION__) L"() Function",
	                        MB_OK,
	                        MAKELANGID(LANG_ENGLISH, SUBLANG_NEUTRAL));

	// https://msdn.microsoft.com/en-us/library/ms688560.aspx

	return niButton == IDOK ? S_OK : S_FALSE;
}

// https://msdn.microsoft.com/en-us/library/ms691457.aspx
//
// REGSVR32.EXE [/S] /U TLS_DEMO.DLL

__declspec(safebuffers)
HRESULT	WINAPI	DllUnregisterServer(VOID)
{
	WCHAR	szOutput[1024];

	// https://msdn.microsoft.com/en-us/library/ms647550.aspx

	INT	niOutput = wsprintf(szOutput, L"Return address @ 0x%p = 0x%p\n",
		                    _AddressOfReturnAddress(), _ReturnAddress());
	INT	niButton = 0;

	if (niOutput <= 0)
		return S_FALSE;

//	szOutput[niOutput] = L'\0';

	// https://msdn.microsoft.com/en-us/library/ms645507.aspx

	niButton = MessageBoxEx(HWND_DESKTOP,
	                        szOutput,
	                        __LPREFIX(__FUNCTION__) L"() Function",
	                        MB_OK,
	                        MAKELANGID(LANG_ENGLISH, SUBLANG_NEUTRAL));

	// https://msdn.microsoft.com/en-us/library/ms688560.aspx

	return niButton == IDOK ? S_OK : S_FALSE;
}

// https://support.microsoft.com/en-us/kb/164787
//
// RUNDLL32.EXE TLS_DEMO.DLL,#1[ <command line>]
// RUNDLL32.EXE TLS_DEMO.DLL,RunDll[ <command line>]

__declspec(safebuffers)
VOID	WINAPI	RunDll(HWND hWindow, HINSTANCE hInstance, LPCSTR lpszCmdLine, INT niCmdShow)
{
	WCHAR	szOutput[1024];

	// https://msdn.microsoft.com/en-us/library/ms647550.aspx

	INT	niOutput = wsprintf(szOutput, L"Return address @ 0x%p = 0x%p\n"
		                              L"\n"
		                              L"Arguments:\n"
		                              L"\tWindow = 0x%p\n"
		                              L"\tInstance = 0x%p\n"
		                              L"\tCmdLine = %s\n"
		                              L"\tCmdShow = %d\n",
		                    _AddressOfReturnAddress(), _ReturnAddress(),
		                    hWindow, hInstance, lpszCmdLine, niCmdShow);

	if (niOutput > 0)
	{
	//	szOutput[niOutput] = L'\0';

		// https://msdn.microsoft.com/en-us/library/ms645507.aspx

		MessageBoxEx(HWND_DESKTOP,
		             szOutput,
		             __LPREFIX(__FUNCTION__) L"() Function",
		             MB_OK,
		             MAKELANGID(LANG_ENGLISH, SUBLANG_NEUTRAL));
	}
}

// https://msdn.microsoft.com/en-us/library/3hxxtd06.aspx
// https://msdn.microsoft.com/en-us/library/988ye33t.aspx
// https://msdn.microsoft.com/en-us/library/ms682583.aspx
// https://msdn.microsoft.com/en-us/library/ms682596.aspx

// https://msdn.microsoft.com/en-us/library/ms679284.aspx

__declspec(safebuffers)
BOOL	WINAPI	_DllMainCRTStartup(HMODULE hModule, DWORD dwReason, CONTEXT *lpContext)
{
	WCHAR	szOutput[1024];

	// https://msdn.microsoft.com/en-us/library/ms647550.aspx

	INT	niOutput = wsprintf(szOutput, L"Return address @ 0x%p = 0x%p\n"
		                              L"\n"
		                              L"Arguments:\n"
		                              L"\tModule = 0x%p\n"
		                              L"\tReason = %u (%s)\n"
		                              L"\tContext = 0x%p\n",
		                    _AddressOfReturnAddress(), _ReturnAddress(),
		                    hModule, dwReason, szReason[dwReason], lpContext);
	INT	niButton = 0;

	if (niOutput <= 0)
		return FALSE;

//	szOutput[niOutput] = L'\0';

	// https://msdn.microsoft.com/en-us/library/ms645507.aspx

	niButton = MessageBoxEx(HWND_DESKTOP,
	                        szOutput,
	                        __LPREFIX(__FUNCTION__) L"() Function",
	                        dwReason == DLL_PROCESS_ATTACH ? MB_OKCANCEL : MB_OK,
	                        MAKELANGID(LANG_ENGLISH, SUBLANG_NEUTRAL));

	if (dwReason != DLL_PROCESS_ATTACH)
		return FALSE;

	return niButton == IDOK;
}

// https://msdn.microsoft.com/en-us/library/f9t8842e.aspx

__declspec(noreturn)
VOID	WINAPI	wWinMainCRTStartup(VOID)
{
	WCHAR	szOutput[1024];
	INT	niOutput;

	// https://msdn.microsoft.com/en-us/library/ms647550.aspx

	niOutput = wsprintf(szOutput, L"Return address @ 0x%p = 0x%p\n"
	                              L"\n"
	                              L"TLS index = %ld\n"
	                              L"TLS value = 0x%p\n"
	                              L"TLS array @ 0x%p\n"
	                              L"TLS block @ 0x%p\n",
	                    _AddressOfReturnAddress(), _ReturnAddress(),
	                    _tls_index,
	                    TlsGetValue(_tls_index),
#ifdef _M_IX86
	                    __readfsdword(44),
	                    ((LPVOID *) __readfsdword(44))[_tls_index]);
#elif _M_AMD64
	                    __readgsqword(88),
	                    ((LPVOID *) __readgsqword(88))[_tls_index]);
#else
#error
#endif
	if (niOutput > 0)
	{
	//	szOutput[niOutput] = L'\0';

		// https://msdn.microsoft.com/en-us/library/ms645507.aspx

		MessageBoxEx(HWND_DESKTOP,
		             szOutput,
		             __LPREFIX(__FUNCTION__) L"() Function",
		             MB_OK,
		             MAKELANGID(LANG_ENGLISH, SUBLANG_NEUTRAL));
	}

	// https://msdn.microsoft.com/en-us/library/ms682658.aspx

	ExitProcess(_tls_index);
}
<<NOKEEP
